/**
 * Mihailescu Maria-Elena
 * grupa 331CC
 * Connect Four pe LED-uri
 */
#define F_CPU 16000000

#include <avr/interrupt.h>
#include <avr/io.h>

#include <util/delay.h>

#include <stdio.h>
#include <stdlib.h>

#include "project_struct.h"

/* if col is different from S, then the game is win by 
 * red if col is R
 * green if col is V
 */
int col = S;

/* left_a  = left_allow  - it allows left  button to be pressed */
/* right_a = right_allow - it allows right button to be pressed*/
/* enter_a = enter_allow - it allows enter button to be pressed*/
int left_a = 1, right_a = 1, enter_a = 1;

/* counter to divide the timer 0 clk - it doesn't allow left_a, right_a and 
 * enter_a to be changed into 1 too fast 
 */
int aux_counter = 0;

ISR(TIMER0_OVF_vect)
{
	aux_counter ++;
	if(aux_counter == 20){
		aux_counter = 0;
		/* allows buttons to be pressed */
		left_a = 1; right_a = 1; enter_a = 1;
		/* deactivates interruption */
		TIMSK0 &= ~(1 << TOIE0);
	}
}

/* timer interruption: compared with OCRA1 */
ISR(TIMER1_COMPA_vect)
{
	/* print time */
	print_nr(time);

	/* decrease time */
	time --;
	int i;

	/* if time has expired */
	if(time == -1){
		time = 9;
		/* simulates enter pressed button */
		matrix[0][place].state = S;
		for(i = 4; i >= 0; i--){
			if(matrix[i][place].state == S){
				matrix[i][place].state = turn;
				break;
			}
		}
		/* change turn */
		if(turn == R)
			turn = V;
		else 
			turn = R;
		/* search if there is a winner */
		col = is_win();
		for(i = 0; i < 5; i ++)
			if(matrix[0][i].state == S){
				place = i;
				break;
			}
		matrix[0][place].state = turn;
	}
}


ISR(PCINT2_vect)
{
	int i;
	/* if left button was pressed and it is allowed to be pressed */
	if((PINC & (1 << LEFT)) == 0 && left_a == 1)
    {
		/* activates interruption */
		TIMSK0 = (1 << TOIE0);
		left_a = 0;
		aux_counter = 0;
        matrix[0][place].state = S;
		for(i = place - 1; i >= 0; i --)
			if(matrix[0][i].state == S){
				place = i;
				break;
			}
		matrix[0][place].state = turn;
    }
    
	/* if right button was pressed and it is allowed to be pressed */
	if((PINC & (1 << RIGHT)) == 0 && right_a == 1)
    {
		/* activates interruption */
		TIMSK0 = (1 << TOIE0);
		right_a = 0;
		aux_counter = 0;
		matrix[0][place].state = S;
		for(i = place + 1; i < 5; i ++){
			if(matrix[0][i].state == S){
				place = i;
				break;
			}
		}
		matrix[0][place].state = turn;
    }
	
	/* if enter button was pressed and it is allowed to be pressed */
	if((PINC & (1 << ENTER)) == 0 && enter_a == 1)
    {
		/* activates interruption */
		TIMSK0 = (1 << TOIE0);
		enter_a = 0;
		aux_counter = 0;
		matrix[0][place].state = S;
		for(i = 4; i >= 0; i--){
			if(matrix[i][place].state == S){
				matrix[i][place].state = turn;
				break;
			}
		}
		if(turn == R)
			turn = V;
		else 
			turn = R;
		time = 9;
		for(i = 0; i < 5; i ++)
			if(matrix[0][i].state == S){
				place = i;
				break;
			}
		col = is_win();
		matrix[0][place].state = turn;
    }
}

/* init timer 0 - timer 0 is used to set left_a, right_a and enter_a 
 * to be pressed - it's a mode to implement debouncing
 */
void timer0_init()
{
    /* prescaler 1024 */
    TCCR0B = (1 << CS00) | (1 << CS02);
}

/* init timer 1  - timer 1 is used to count the seconds for the display */
void timer1_init()
{
    /* for 1 second */
    OCR1A = 15625;
    
    /* activates interruption */
    TIMSK1 = (1 << OCIE1A) | (1 << TOIE1);
    
    /* CTC mode with OCR1A */
    /* prescaler 1024 */
    TCCR1B = (5 << CS10) | (1 << WGM12);
}

/* print matrix */
void print_matrix()
{
	int i, j;
	for(j = 0; j < 5; j ++){
			PORTB = 0;
			PORTD &= ~((1 << PD0) | (1 << PD1) | (1 << PD2) | (1 << PD3) 
				| (1 << PD4) | (1 << PD5) | (1 << PD7));
			PORTD |= (1 << matrix[0][j].gnd);
			for(i = 0; i < 5; i ++){
				if(matrix[i][j].state == R){
					if(i < 3)
						PORTB |= (1 << matrix[i][j].red);
					else
						PORTD |= (1 << matrix[i][j].red);
				} else if(matrix[i][j].state == V){
					PORTB |= (1 << matrix[i][j].green);
				} else{
					if(i < 3)
						PORTB &= ~(1 << matrix[i][j].red);
					else
						PORTD &= ~(1 << matrix[i][j].red);
					PORTB &= ~(1 << matrix[i][j].green);
				}
				_delay_ms(1);
			}
	}
}

/* function for the winning phase */
void win(int mode)
{
	int i, j;
	/* deactivate interruptions */
	/* time is freezed and buttons do nothing */
	cli();
	while(1){
		PORTB = 0;
		PORTD &= ~((1 << PD0) | (1 << PD1) | (1 << PD2) | (1 << PD3) 
				| (1 << PD4) | (1 << PD5) | (1 << PD7));
		PORTD |= (1 << matrix[0][0].gnd) | (1 << matrix[0][1].gnd)
				| (1 << matrix[0][2].gnd) | (1 << matrix[0][3].gnd)
				| (1 << matrix[0][4].gnd);
		/* blinking matrix in the color of the winner */
		if(mode == R){
			for(i = 0; i < 5; i++)
				for(j = 0; j < 5; j++)
					if(i < 3)
						PORTB |= (1 << matrix[i][j].red);
					else
						PORTD |= (1 << matrix[i][j].red);
					
		} else{
			for(i = 0; i < 5; i++)
				for(j = 0; j < 5; j++)
					PORTB |= (1 << matrix[i][j].green);
					
		}
		_delay_ms(250);
		for(i = 0; i < 25; i++)
			print_matrix();
	}
}

/* game logic */
void connect_four()
{
	PORTB |= (1 << PB0);
	int i, j;

	while(1)
	{
		/* print matrix */
		print_matrix();
		/* see if there is a winner of the game */
		if(col != S)
			win(col);
		if(col == V)
			for(i = 0; i < 5; i++)
				for(j = 0; j < 5; j++)
					if(matrix[i][j].state == R)
						matrix[i][j].state = S;
		if(col == R)
			for(i = 0; i < 5; i++)
				for(j = 0; j < 5; j++)
					if(matrix[i][j].state == V)
						matrix[i][j].state = S;
	}

}

int main(void)
{
	init();
	IO_init();
	/* activates interruptions */
	sei();
	timer1_init();
	timer0_init();
	connect_four();
	return 0;
}